"""Module containing classes and functions used for styling plotly figures.

Contains:
- LineStyle: Dataclass containing information on how to style a line on a line
    chart.
"""
from .line_style import LineStyle
